import _queue
import logging
import sys
import traceback

from pymavlink.mavutil import mavlink
from threading import Thread
from multiprocessing import Queue

from pynanopix import ConnectionManager
from pynanopix.MultiprocessLogger import MultiprocessLogger, LoggerMessage
from pynanopix.utils import Utils


class MavlinkConnector:
    def __init__(self, conn_manager: ConnectionManager):

        self._all_msgs_listeners = []
        self._conn_manager = conn_manager
        self._mav = Utils.get_mavlink_instance()
        self._tx_msg_queue = Queue() # tuples: (target_id, mav message to tx)
        self._rx_msg_queue = Queue() # received messages
        self._log_queue = MultiprocessLogger.get_queue()
        self._is_running = False
        self._thread = Thread(target=self._main_routine)

    def add_listener(self, all_msgs_listener):
        self._all_msgs_listeners.append(all_msgs_listener)

    def get_tx_queue(self) -> Queue:
        return self._tx_msg_queue

    def _main_routine(self):
        self._log_queue.put(LoggerMessage("MavlinkConnector main routine started", logging.INFO))
        conn_rx_queue = self._conn_manager.get_rx_queue()
        while True:

            ###################### Sending: ###########################
            while self._tx_msg_queue.qsize():
                tx_msg = None
                target_id = None
                try:
                    target_id, tx_msg = self._tx_msg_queue.get(False)
                except _queue.Empty:
                    break

                if tx_msg is not None:
                    try:
                        self._mav.srcSystem = target_id
                        self._mav.srcComponent = 0
                        tx_msg.pack(self._mav)
                        b = tx_msg.get_msgbuf()
                        bytes_buffer = []
                        for c in b:
                            bytes_buffer.append(int(c))

                        self._conn_manager.tx_data(bytes_buffer)
                    except Exception as e:
                        exc_type, exc_value, exc_traceback = sys.exc_info()
                        trace_str = repr(traceback.format_exception(exc_type, exc_value, exc_traceback))
                        self._log_queue.put(LoggerMessage(
                            "MavlinkConnector send msg FAILED: " + trace_str, logging.ERROR))

            ###################### Receiving: ###########################
            while conn_rx_queue.qsize() > 0:
                rx_data = None
                try:
                    rx_data = conn_rx_queue.get(False)
                except _queue.Empty:
                    break

                if rx_data is not None:
                    msgs = self._mav.parse_buffer(rx_data)
                    if msgs is not None:
                        for msg in msgs:
                            for lsnr in self._all_msgs_listeners:
                                lsnr.on_mav_packet_received(msg)

            if not self._is_running and conn_rx_queue.qsize() == 0 and self._tx_msg_queue.qsize() == 0:
                break
        self._log_queue.put(LoggerMessage("MavlinkConnector main routine finished", logging.INFO))

    def start(self):
        self._is_running = True
        self._thread.start()

    def stop(self):
        self._is_running = False
        self._thread.join()
