import threading
from math import pi
from multiprocessing import Queue
from time import time, sleep

from pymavlink.dialects.v10.ardupilotmega import MAV_CMD_NAV_TAKEOFF, MAV_CMD_NAV_LAND, MAV_CMD_COMPONENT_ARM_DISARM, \
    MAVLINK_MSG_ID_LOCAL_POSITION_NED, MAV_FRAME_LOCAL_OFFSET_NED, MAV_FRAME_LOCAL_NED, MAV_CMD_CONDITION_YAW, \
    MAVLINK_MSG_ID_RAW_IMU, MAVLINK_MSG_ID_ATTITUDE
from pymavlink.mavutil import mavlink
from pynanopix.utils import Utils


class NanopixController:
    ATTITUDE_TELEMETRY_STREAM_ID = 3
    LOCAL_POS_TELEMETRY_STREAM_ID = 5
    DEFAULT_POS_SPEED = 0.4
    DEFAULT_YAW_SPEED = 35
    DEFAULT_POS_Z_MOVE_SPEED = 0.2

    def __init__(self, sys_index, tx_msg_queue: Queue):
        self._sys_index = sys_index
        self._tx_msg_queue = tx_msg_queue
        self._bat_voltage = 0
        self._mav = Utils.get_mavlink_instance()
        self._rc_roll = 1500
        self._rc_pitch = 1500
        self._rc_throttle = 1500
        self._rc_yaw = 1500
        self._is_manual_control_enabled = False
        self._is_local_pos_received = False
        self._x_pos = None
        self._y_pos = None
        self._z_pos = None
        self._is_attitude_received = False
        self._roll = None
        self._pitch = None
        self._yaw = None
        self._roll_rate = None
        self._pitch_rate = None
        self._yaw_rate = None
        self._disarm_repeats = 0
        self._arm_repeats = 0
        self._takeoff_repeats = 0
        self._land_repeats = 0
        self._front_led_repeats = 0
        self._back_led_repeats = 0
        self._start_ts = time()
        self._print_statustext = False

    def _start(self):
        self._is_local_pos_received = False
        self._is_attitude_received = False
        self._is_manual_control_enabled = True
        #self._manual_controller_task()

    def _stop(self):
        self._is_local_pos_received = True
        self._is_attitude_received = True
        self._is_manual_control_enabled = False

    def _manual_controller_task(self):
        mav_msg = self._mav.manual_control_encode(0, self._rc_roll, self._rc_pitch, self._rc_throttle, self._rc_yaw, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))
        if self._is_manual_control_enabled:
            threading.Timer(0.015, self._manual_controller_task).start()

    def _send_localpos_telem_request(self, stream_rate):
        mav_msg = self._mav.request_data_stream_encode(0, 0, NanopixController.LOCAL_POS_TELEMETRY_STREAM_ID,
                                                       stream_rate, 1)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def _send_attitude_telem_request(self, stream_rate):
        mav_msg = self._mav.request_data_stream_encode(0, 0, NanopixController.ATTITUDE_TELEMETRY_STREAM_ID,
                                                       stream_rate, 1)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def _send_mav_packet(self, msg):
        self._tx_msg_queue.put((self._sys_index, msg))

    def _parse_incoming_message(self, mav_msg):
        if mav_msg.id == mavlink.MAVLINK_MSG_ID_SYS_STATUS:
            self._bat_voltage = mav_msg.voltage_battery / 1000.0
            #print("Nanopix %d bat voltage: %f" % (self._sys_index, self._bat_voltage))
        if mav_msg.id == mavlink.MAVLINK_MSG_ID_STATUSTEXT:
            if self._print_statustext:
                print("Nanopix %d log: %s" % (self._sys_index, mav_msg.text))
        if mav_msg.id == MAVLINK_MSG_ID_LOCAL_POSITION_NED:
            self._x_pos = mav_msg.x
            self._y_pos = mav_msg.y
            self._z_pos = mav_msg.z
            self._is_local_pos_received = True
            #print("Nanopix %d pos: %f %f %f" % (self._sys_index, mav_msg.x, mav_msg.y, mav_msg.z))
        if mav_msg.id == MAVLINK_MSG_ID_ATTITUDE:
            self._roll = mav_msg.roll
            self._pitch = mav_msg.pitch
            self._yaw = mav_msg.yaw
            self._roll_rate = mav_msg.rollspeed
            self._pitch_rate = mav_msg.pitchspeed
            self._yaw_rate = mav_msg.yawspeed
            self._is_attitude_received = True

    def set_log_printing(self, is_enabled):
        self._print_statustext = is_enabled

    def set_roll(self, roll_deg, roll_only: bool = True): # roll_value in degrees [-50, 50]
        self._rc_roll = 1500 + roll_deg * 10
        if roll_only:
            self._rc_pitch = 1500
            self._rc_yaw = 1500
            self._rc_throttle = 1500

    def set_pitch(self, pitch_deg, pitch_only: bool = True): # pitch_value in degrees [-50, 50]
        self._rc_pitch = 1500 + pitch_deg * 10
        if pitch_only:
            self._rc_roll = 1500
            self._rc_yaw = 1500
            self._rc_throttle = 1500

    def set_yaw_speed(self, speed_deg_sec):  # yaw_speed in degrees/sec [-50, 50]
        self._rc_yaw = 1500 + speed_deg_sec * 10

    def set_altitude_speed(self, altitude_speed):  # altitude_speed in meters/sec [-0.5, 0.5]
        self._rc_throttle = 1500 + altitude_speed * 1000

    def _set_color(self, led_name, R, G, B):
        if R < 0:
            R = 0
        if R > 255:
            R = 255
        if G < 0:
            G = 0
        if G > 255:
            G = 255
        if B < 0:
            B = 0
        if B > 255:
            B = 255
        color = int.from_bytes([0, int(R), int(G), int(B)], byteorder='big', signed=False)
        mav_msg = self._mav.param_set_encode(0, 0, led_name, Utils.float_from_integer(color), 9)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def set_front_led(self, R, G, B):
        self._set_color(b'FrontLed', R, G, B)

        if self._front_led_repeats == 0:
            self._front_led_repeats = 3
        if self._front_led_repeats > 0:
            self._front_led_repeats -= 1
            if self._front_led_repeats > 0:
                threading.Timer(0.01, self.set_front_led, args=[R, G, B]).start()

    def set_back_led(self, R, G, B):
        self._set_color(b'BackLed', R, G, B)

        if self._back_led_repeats == 0:
            self._back_led_repeats = 3
        if self._back_led_repeats > 0:
            self._back_led_repeats -= 1
            if self._back_led_repeats > 0:
                threading.Timer(0.01, self.set_back_led, args=[R, G, B]).start()

    def takeoff(self):
        mav_msg = self._mav.command_long_encode(0, 0, MAV_CMD_NAV_TAKEOFF, 0, 0, 0, 0, 0, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

        if self._takeoff_repeats == 0:
            self._takeoff_repeats = 5
        if self._takeoff_repeats > 0:
            self._takeoff_repeats -= 1
            if self._takeoff_repeats > 0:
                threading.Timer(0.05, self.takeoff).start()

    def land(self):
        mav_msg = self._mav.command_long_encode(0, 0, MAV_CMD_NAV_LAND, 0, 0, 0, 0, 0, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

        if self._land_repeats == 0:
            self._land_repeats = 5
        if self._land_repeats > 0:
            self._land_repeats -= 1
            if self._land_repeats > 0:
                threading.Timer(0.05, self.land).start()

    def arm_motors(self):
        mav_msg = self._mav.command_long_encode(0, 0, MAV_CMD_COMPONENT_ARM_DISARM, 0, 1, 0, 0, 0, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

        if self._arm_repeats == 0:
            self._arm_repeats = 5
        if self._arm_repeats > 0:
            self._arm_repeats -= 1
            if self._arm_repeats > 0:
                threading.Timer(0.05, self.arm_motors).start()

    def disarm_motors(self):
        mav_msg = self._mav.command_long_encode(0, 0, MAV_CMD_COMPONENT_ARM_DISARM, 0, 0, 0, 0, 0, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

        if self._disarm_repeats == 0:
            self._disarm_repeats = 5
        if self._disarm_repeats > 0:
            self._disarm_repeats -= 1
            if self._disarm_repeats > 0:
                threading.Timer(0.05, self.disarm_motors).start()

    def _enable_and_wait_local_pos_telem(self):
        timeout = 1.2
        stream_rate = 25
        while not self._is_local_pos_received and timeout > 0:
            self._send_localpos_telem_request(stream_rate)
            timeout -= 0.01
            sleep(0.05)

    def _enable_and_wait_attitude_telem(self):
        timeout = 1.2
        stream_rate = 25
        while not self._is_attitude_received and timeout > 0:
            self._send_attitude_telem_request(stream_rate)
            timeout -= 0.01
            sleep(0.05)

    def get_pos_x(self) -> float:
        self._enable_and_wait_local_pos_telem()
        return self._x_pos

    def get_pos_y(self) -> float:
        self._enable_and_wait_local_pos_telem()
        return self._y_pos

    def get_pos_z(self) -> float:
        self._enable_and_wait_local_pos_telem()
        return self._z_pos

    def get_roll(self) -> float:
        self._enable_and_wait_attitude_telem()
        return (self._roll / pi) * 180

    def get_pitch(self) -> float:
        self._enable_and_wait_attitude_telem()
        return (self._pitch / pi) * 180

    def get_yaw(self) -> float:
        self._enable_and_wait_attitude_telem()
        return (self._yaw / pi) * 180

    def _send_pos(self, x, y, z, vx, vy, vz, is_relative):
        if is_relative:
            frame = MAV_FRAME_LOCAL_OFFSET_NED
        else:
            frame = MAV_FRAME_LOCAL_NED
        mav_msg = self._mav.set_position_target_local_ned_encode(
            (int)((time() - self._start_ts) * 1000), 0, 0, frame, 0, x, y, z, vx, vy, vz, 0, 0, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def _send_z_pos(self, z, is_relative):
        self._send_pos(0, 0, z, 0, 0, NanopixController.DEFAULT_POS_Z_MOVE_SPEED, is_relative)

    def _send_xy_pos(self, x, y, is_relative):
        self._send_pos(x, y, 0, NanopixController.DEFAULT_POS_SPEED,
                       NanopixController.DEFAULT_POS_SPEED, 0, is_relative)

    def set_xy_pos(self, x_meters, y_meters):
        for i in range(0, 2):
            self._send_xy_pos(x_meters, y_meters, False)
            sleep(0.005)

    def offset_xy_pos(self, dx_meters, dy_meters):
        self._send_xy_pos(dx_meters, dy_meters, True)

    def offset_altitude(self, delta_meters):
        self._send_z_pos(delta_meters, True)

    def set_altitude(self, z_pos):
        self._send_z_pos(z_pos, False)

    def move_forward(self, meters):
        self.offset_xy_pos(0, meters)

    def move_backward(self, meters):
        self.offset_xy_pos(0, -meters)

    def move_right(self, meters):
        self.offset_xy_pos(meters, 0)

    def move_left(self, meters):
        self.offset_xy_pos(-meters, 0)

    def _set_yaw(self, yaw_value, is_relative):
        mav_msg = self._mav.command_int_encode(0, 0, 0, MAV_CMD_CONDITION_YAW, 0, 0, yaw_value,
                                     NanopixController.DEFAULT_YAW_SPEED, 1, is_relative, 0, 0, 0)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def rotate_yaw(self, yaw_offset_deg):
        self._set_yaw(yaw_offset_deg, 1)

    def set_yaw(self, yaw_value_deg):
        self._set_yaw(yaw_value_deg, 0)

    def _set_headfree(self, value):
        mav_msg = self._mav.param_set_encode(0, 0, b"Headfree", value, 9)
        self._tx_msg_queue.put((self._sys_index, mav_msg))

    def enable_headfree(self):
        self._set_headfree(1)

    def disable_headfree(self):
        self._set_headfree(0)