import logging
from time import time

from pymavlink.mavutil import mavlink

from pynanopix.ConnectionManager import UdpConnectionManager
from pynanopix.MavlinkConnector import MavlinkConnector
from pynanopix.MultiprocessLogger import MultiprocessLogger
from pynanopix.NanopixController import NanopixController


class NanopixManager:
    _instance = None
    def __init__(self):
        self._mp_logger = MultiprocessLogger("logs.txt", logging.DEBUG, 2, 60)
        self._mp_logger.start()
        
        self._controllers = {}
        self._sys_status_received = False

        self._conn = UdpConnectionManager()
        self._conn.start()

        self._mav_conn = MavlinkConnector(self._conn)
        self._mav_conn.add_listener(self)
        self._mav_conn.start()

    @staticmethod
    def getInstance():
        if NanopixManager._instance is None:
            NanopixManager._instance = NanopixManager()
        return NanopixManager._instance

    def join_nanopixes(self) -> []:
        join_time = 3
        join_start = None
        join_start_timeout = 10
        start_ts = time()
        while True:
            if join_start is None:
                if time() - start_ts > join_start_timeout:
                    return []

                if self._sys_status_received:
                    join_start = time()
            else:
                if time() - join_start > join_time:
                    break

        res_controllers = []
        for key in self._controllers:
            res_controllers.append(self._controllers[key])

        if self._sys_status_received and len(self._controllers) == 0: # Wi-Fi case
            np_ctrl = NanopixController(0, self._mav_conn.get_tx_queue())
            res_controllers.append(np_ctrl)
            self._controllers[0] = np_ctrl

        for key in self._controllers:
            self._controllers[key]._start()

        if len(res_controllers) == 0:
            print("Connect PC to Nanopix via Wi-Fi")
            NanopixManager.getInstance().stop()
            exit(0)

        return res_controllers

    def on_mav_packet_received(self, mav_msg):
        #print(type(mav_msg))
        controller_state_msg = False
        if mav_msg.id == mavlink.MAVLINK_MSG_ID_SYS_STATUS:
            self._sys_status_received = True
            if mav_msg.get_srcSystem() == 255 and mav_msg.get_srcComponent() == 255:
                controller_state_msg = True
                peers = mav_msg.onboard_control_sensors_present
                for i in range(0, 20):
                    if 1 << i & peers:
                        if i not in self._controllers:
                            self._controllers[i] = NanopixController(i, self._mav_conn.get_tx_queue())
                    else:
                        if i in self._controllers:
                            self._controllers[i]._stop()
                            del self._controllers[i]

        if not controller_state_msg:
            sys_id = mav_msg.get_srcSystem()
            if sys_id in self._controllers:
                self._controllers[sys_id]._parse_incoming_message(mav_msg)

    def stop(self):
        for key in self._controllers:
            self._controllers[key]._stop()
        self._conn.stop()
        self._mav_conn.stop()
        self._mp_logger.stop()



